package Plugins::SmartMix::Mixer;

use strict;

use Scalar::Util qw(blessed);
use Tie::Cache::LRU;
use URI::Escape qw(uri_unescape);

use Slim::Utils::Log;
use Slim::Utils::Prefs;
use Slim::Utils::Strings qw(cstring);

use Plugins::SmartMix::API;
use Plugins::SmartMix::Services;

use constant MAX_OLD_TRACKS => 3;		# maximum number of played tracks to keep in the playlist
use constant MIN_NEW_TRACKS => 4;		# minimum number of tracks to be queued up
use constant NEW_TRACKS     => 5;		# number of tracks to fetch at once - use maximum to limit hit rate

# keep a small list of urls used in the playlist
tie my %urlCache, 'Tie::Cache::LRU', (MIN_NEW_TRACKS + MAX_OLD_TRACKS) * 2;
%urlCache = ();

my $prefs = preferences('plugin.smartmix');
my $log = logger('plugin.smartmix');

# playlist commands that will stop a "radio"
my %stopcommands = (
	'clear'	     => 1,
	'loadtracks' => 1, # multiple play
	'playtracks' => 1, # single play
	'load'       => 1, # old style url load (no play)
	'play'       => 1, # old style url play
	'loadalbum'  => 1, # old style multi-item load
	'playalbum'  => 1, # old style multi-item play
);

sub init {
	Slim::Control::Request::addDispatch(['smartmix', '_mix'], [1, 0, 1, \&cliCreateMix]);

	# listen to playlist change events so we know when our own playlist ends
	Slim::Control::Request::subscribe(\&onPlaylistChange, [['playlist'], ['newsong', 'delete', keys %stopcommands]]);
	Slim::Control::Request::subscribe(\&onSkip, [['playlist'], ['jump']]) if $Plugins::SmartMix::Plugin::FEEDBACK;
}

sub createMixFromCurrentTrack {
	my ($client, $cb, $params) = @_;
	
	my $items = [];
	my $title;
	
	if ( $client && $client->playingSong && (my $track = $client->playingSong->track) ) {
		my ($artist, $title, $duration) = getMixablePropertiesFromTrack($client, $track);
		
		# don't seed from radio stations - only do if we're playing from some track based source
		if ($artist && $title) {
			main::INFOLOG && $log->info("Auto-mixing from current track in playlist: $title - $artist");
			
			createMix($client, $cb, undef, {
				track_id => getTrackId($track),
				title    => $title,
				artist   => $artist,
			}, $duration, $duration);
			
			# resume playback if not playing
			$client->controller->resume() if $client->controller->isPaused;
			$client->controller->play() if $client->controller->isStopped;
			
			return;
		}
	}
	
	if (!scalar @$items) {
		push @$items, {
			name => cstring($client, 'PLUGIN_SMARTMIX_PLAYLIST_FAILURE', $title),
			showBriefly => 1,
		};
	}
	
	$cb->({
		items => $items
	}) if $cb;
}

sub createMixFromTopArtists {
	my ($client, $cb, $params) = @_;
	
	my $dbh = Slim::Schema->dbh;
	
	my $ignoreGenres = '';
	
	if ( my $genres = $prefs->get('ignoreGenres') ) {
		
		my $sql = "SELECT id FROM genres WHERE " . join(' OR ', map {
			"name LIKE '$_%'"
		} split /\s*,\s*/, $genres);
		
		my $genres_ref = $dbh->selectall_arrayref($sql);
		
		if ( scalar @$genres_ref ) {
			$ignoreGenres = ' AND tracks.id = genre_track.track AND NOT genre_track.genre IN (' . join(', ', map { $_->[0] } @$genres_ref) . ')';
		}
	}

	my $sql = qq{
		SELECT contributors.name, SUM( persistent.playcount ) AS playcount
		FROM tracks, contributors LEFT JOIN tracks_persistent persistent ON persistent.urlmd5 = tracks.urlmd5 
	} . ($ignoreGenres ? ', genre_track' : '') . qq{
		WHERE playcount > 0 AND contributors.id = tracks.primary_artist $ignoreGenres 
		GROUP BY tracks.primary_artist 
		ORDER BY playcount DESC
		LIMIT 5
	};

	my $artists_ref = $dbh->selectall_arrayref($sql);
	
	main::INFOLOG && $log->is_info && $log->info('Top 5 artists according to playcount: ' . Data::Dump::dump($artists_ref));

	createMix($client, $cb, undef, {
		title  => cstring($client, 'PLUGIN_SMARTMIX_TOP_ARTISTS'),
		artist => [ map { $_->[0] } @$artists_ref ],
		type => 'artist-radio',
		variety => 0.5,
		distribution => 'wandering',
		adventurousness => 0.1,
	});
}

sub cliCreateMix {
	my $request = shift;
  	
	my $client  = $request->client();
	my $mix     = lc( uri_unescape($request->getParam('_mix') || '') );

	my ($recipe) = grep { lc($_->{name}) eq $mix } @{ Plugins::SmartMix::Plugin::getRecipes() };

	if ($client && $mix eq '_top_artists') {
		createMixFromTopArtists($client);
	}
	elsif ($client && $mix eq '_recommended') {
		createMix($client, undef, undef, {
			name => cstring($client, 'PLUGIN_SMARTMIX_RECOMMENDED'),
			type => 'catalog-radio',
		});
	}
	elsif (!$recipe || !$client) {
		$request->setStatusBadParams();
		return;
	}
	else {
		createMix($client, undef, undef, $recipe);
	}
	
	$request->setStatusDone();
}

sub createMix {
	# $params - never used, but passed when called from a menu
	my ($client, $cb, $params, $args, $addToQueue, $dontZap) = @_;

	# Slim::BUttons::XMLBrowser is adding a time or something as an additional passthrough parameter - no idea where this is coming from
	$addToQueue = undef if $params->{isButton} && $addToQueue =~ /:/;

	# XXX - should we wipe the feedback queue here?
	#$prefs->client($client)->remove('feedback');
	
	my $catalog = $prefs->get('songCatalogId');
	
	# if seed_catalog is being used, make sure it's set to the user's catalog
	$args->{seed_catalog} = $catalog if $args->{seed_catalog} || $args->{type} =~ /^catalog/i;
	
	my $createMixCb = sub {
		my ($args, $title) = @_;

		# delete existing session
		clearMixSession($client);
		
		$args->{bucket} = getBuckets($client);
		
		Plugins::SmartMix::API->playlistDynamicCreate($args, sub {
			_createMixCb($client, shift, $title, $cb, $addToQueue, $dontZap)
		});
	};
	
	my $title = delete($args->{title}) || (ref $args->{artist} eq 'ARRAY' ? $args->{artist}->[0] : $args->{artist}) || cstring($client, 'PLUGIN_SMARTMIX_SMART');
	
	main::INFOLOG && $log->is_info && $log->info("Starting dynamic mix using: " . Data::Dump::dump($args));
	
	if ($args->{track_id}) {
		if (ref $args->{track_id} eq 'ARRAY') {
			$args->{track_id} = [ map { /:/ ? $_ : "$catalog:song:$_" } @{$args->{track_id}} ];
		}
		elsif ($args->{track_id} !~ /:/) {
			$args->{track_id} = $catalog . ':song:' . $args->{track_id};
		}
		
		$args = {
			type    => 'song-radio',
			song_id => $args->{track_id},
			distribution => 'wandering',
			variety => 0.8,
		};

		$createMixCb->($args, $title);
	}
	elsif ($args && $args->{name}) {
		$createMixCb->($args, delete $args->{name})
	}
	elsif ($args->{artist}) {
		$args = {
			type   => $args->{type} =~ /artist/i ? $args->{type} : 'artist-radio',
			artist => $args->{artist},
			distribution => 'wandering',
			variety => 0.8,
		};
		
		$createMixCb->($args, $title);
	}
	elsif ($args->{type} && ($args->{artist} || $args->{style} || $args->{description} || $args->{mood})) {
		$createMixCb->($args);
	}
	else {
		main::INFOLOG && $log->info("Should we really play this configuration? " . Data::Dump::dump($args));
		$createMixCb->($args);
	}
	
	$prefs->client($client->master)->set('recipe', $args);
}

sub onPlaylistChange {
	my $request = shift;
	my $client  = $request->client();

	return if !defined $client;
	$client = $client->master;
	return if $request->source && $request->source eq 'PLUGIN_SMARTMIX';
	return if !$prefs->client($client)->get('session_id') && !$prefs->client($client)->get('neverStopTheMusic');

	# don't interfere with the automatically adding RandomPlay and SugarCube plugins
	# stop smart mixing when a new RandomPlay mode is started or SugarCube is at work
	if (
		( Slim::Utils::PluginManager->isEnabled('Slim::Plugin::RandomPlay::Plugin') && Slim::Plugin::RandomPlay::Plugin::active($client) )
		|| ( Slim::Utils::PluginManager->isEnabled('Plugins::SugarCube::Plugin') && preferences('plugin.SugarCube')->client($client)->get('sugarcube_status') )
	) {
	
		if ($prefs->client($client)->get('session_id')) {
			main::INFOLOG && $log->is_info && $log->info('Cyclic mode ending as some random mix is playing');
			clearMixSession($client);
		}
		
		return;
	}

	my $songIndex = Slim::Player::Source::streamingSongIndex($client) || 0;
	
	if ( main::INFOLOG && $log->is_info ) {
		$log->info(sprintf("Received command %s", $request->getRequestString));
		$log->info(sprintf("While in mode: %s, from %s", ($prefs->client($client)->get('session_id') || 'new!'), $client->name));
	}

	if ( $request->isCommand( [['playlist'], ['newsong', 'delete']] ) ) {
		
		if ( $prefs->client($client)->get('session_id') ) {

			# make sure we don't have any unwanted tracks, samples etc. in the queue
			my $playlist  = Slim::Player::Playlist::playList($client);
			my $remaining = [ @$playlist[$songIndex..$#$playlist] ];
			
			foreach my $track ( @$remaining ) {

				next if $urlCache{blessed($track) ? $track->url : $track};
				
				if ( Plugins::SmartMix::Services->skipTrack($client, $track) ) {
					my $request = $client->execute(['playlist', 'deleteitem', $track]);
					$request->source('PLUGIN_SMARTMIX');
					
					main::INFOLOG && $log->info("Removing track from queue: " . $track->url);
				}
				else {
					$urlCache{$track->url}++;
				}
			}

			my $songsToKeep = $prefs->get('oldtracks') || MAX_OLD_TRACKS;
			my $playlistCount = Slim::Player::Playlist::count($client);
	
			if ($songIndex && $songsToKeep && $songIndex > $songsToKeep) {
	
				main::INFOLOG && $log->is_info && $log->info( sprintf("Stripping off %i completed track(s)", $songIndex - $songsToKeep) );
	
				# Delete tracks before this one on the playlist
				for (my $i = 0; $i < $songIndex - $songsToKeep; $i++) {
					my $request = $client->execute(['playlist', 'delete', 0]);
					$request->source('PLUGIN_SMARTMIX');
				}
			}
	
			addTracks($client);
		}
		# create mix based on last track if we near the end, repeat is off and neverStopTheMusic is set
		elsif ( !Slim::Player::Playlist::repeat($client) ) {
			neverStopTheMusic($client);
		}

	} 
	elsif ( $request->isCommand( [['playlist'], [keys %stopcommands]] ) && $request->getParam('_item') !~ /^smartmix:/ ) {

		main::INFOLOG && $log->is_info && $log->info( sprintf("Cyclic mode ending due to playlist: %s command", $request->getRequestString) );

		# we might want to keep the session around to continue at a later stage?
		clearMixSession($client);
	}
}

sub _createMixCb {
	my ($client, $result, $title, $cb, $addToQueue, $dontZap) = @_;
	
	my $msg = cstring($client, 'PLUGIN_SMARTMIX_PLAYING', $title);
	
	if ( $result->{response} && (my $sid = $result->{response}->{session_id}) ) {
		main::INFOLOG && $log->is_info && $log->info("Got dynamic playlist session_id: $sid");
		
		my $clientPrefs = $prefs->client($client->master);

		$clientPrefs->set('session_id', $sid);
		$clientPrefs->set('mixing', $title);
		
		addTracks($client, $addToQueue ? undef : 'play', $addToQueue && !$dontZap ? 'zap' : undef);
	}
	else {
		$msg = cstring($client, 'PLUGIN_SMARTMIX_PLAYLIST_FAILURE', $title);
		$log->warn($msg);
		main::INFOLOG && $log->is_info && $log->info(Data::Dump::dump($result));
	}

	$cb->({
		items => [{
			name => $msg,
			showBriefly => 1,
			nowPlaying  => 1, # then return to Now Playing
		}]
	}) if $cb;
}

sub getBuckets {
	my $client = shift;
	
	my $buckets = [];

	if ( !$prefs->client($client)->get('localOnly') && (my $services = Plugins::SmartMix::Services->getServices($client)) ) {
		push @$buckets, @$services;
	}

	push @$buckets, 'id:' . $prefs->get('songCatalogId') if !$prefs->get('disable_local') || !scalar @$buckets;

	return $buckets;
}

sub clearMixSession {
	my ($client) = @_;

	my $clientPrefs = $prefs->client($client->master);
		
	if ($clientPrefs) {
		$clientPrefs->remove('session_id');
		$clientPrefs->remove('mixing');
		$clientPrefs->remove('recipe');
	}
}

sub _mergeRecipes {
	my ($client, $new) = @_;
	
	return unless $client;
	$client = $client->master;
	
	my $recipe = $prefs->client($client)->get('recipe') || {};
	return $recipe unless $new;
	
	$new ||= {};
	
	while (my ($k, $v) = each %$new) {
		next if ref $v;		# can't deal with lists yet
		next if $k eq 'name';
		
		# some "steering" commands can't be used to create a mix
		next if $k =~ /^(target_|more_|less_)/;

		# "steering" parameters use different format than playlist creation for some min/max values
		$k =~ s/^(max|min)_(song|artist)_/$2_$1_/;
		
		$recipe->{$k} = $v;
	}

	$prefs->client($client)->set('recipe', $recipe);

	# XXX - legacy? Shouldn't we be using song_id instead of track_id to start with?
	delete $recipe->{name};
	$recipe->{song_id} = delete $recipe->{track_id} if $recipe->{track_id};
	
	return $recipe;
}

sub steer {
	my ($client, $cb, $params, $args) = @_;

	my $controlCb = sub {
		my $params = shift;

		_mergeRecipes($client, $params);

		$params->{session_id} ||= $prefs->client($client->master)->get('session_id');
		Plugins::SmartMix::API->playlistDynamicSteer($params, sub {
			$cb->({
				items => [{
					name => cstring($client, 'PLUGIN_SMARTMIX_OPTIONS_SET'),
					showBriefly => 1,
				}]
			}) if $cb;
		});
	};
	
	if ( my $trackId = ($args->{more_like_this} || $args->{less_like_this}) ) {
		Plugins::SmartMix::API->getSongId($trackId, sub {
			my $id = shift;
			
			if ($id) {
				$args->{more_like_this} = $id if $args->{more_like_this};
				$args->{less_like_this} = $id if $args->{less_like_this};

				$controlCb->($args);
			}
			else {
				main::INFOLOG && $log->is_info && $log->info('no track id was found for ' . $trackId);
				$cb->({
					items => [{
						name => cstring($client, 'PLUGIN_SMARTMIX_OPTIONS_FAILURE'),
						showBriefly => 1,
					}]
				}) if $cb;
			}
		});
	}
	elsif ($args) {
		$controlCb->($args);
	}
	else {
		main::INFOLOG && $log->is_info && $log->info('no control options were found');
		$cb->({
			items => [{
				name => cstring($client, 'PLUGIN_SMARTMIX_OPTIONS_FAILURE'),
				showBriefly => 1,
			}]
		}) if $cb;
	}
}

# Valid feedback keys:
# http://developer.echonest.com/docs/v4/standard.html#dynamic-feedback
sub feedback {
	my ($client, $cb, $params, $args) = @_;
	
	my $fb = $prefs->client($client)->get('feedback') || {};
	
	while ( my ($k, $v) = each( %{$args} ) ) {
		# we only support a sub-set of feedback commands for now
		next unless $k =~ /ban_artist|skip_song/;
		next unless $v;
		
		if ( my $id = Plugins::SmartMix::API->getSongId($v) ) {
			$fb->{$k} ||= [];
			push @{$fb->{$k}}, $id unless grep /$id/i, @{$fb->{$k}};
		}
		
	}
	
	$prefs->client($client)->set('feedback', $fb);
	
	main::INFOLOG && $log->is_info && $log->info('Updated session feedback: ' . Data::Dump::dump($fb));

	$cb->({
		items => [{
			name => cstring($client, 'PLUGIN_SMARTMIX_OPTIONS_SET'),
			showBriefly => 1,
		}]
	}) if $cb;
}

sub restartMix {
	my ($client, $cb, $params, $args) = @_;

	my $recipe = _mergeRecipes($client, $args);
	
	$recipe = Plugins::SmartMix::Plugin::removeCatalogIdFromRecipe($recipe);
	$recipe->{bucket} = getBuckets($client);

	$recipe->{session_id} ||= $prefs->client($client->master)->get('session_id');
	
	Plugins::SmartMix::API->playlistDynamicRestart($recipe, sub {
		$cb->({
			items => [{
				name => cstring($client, 'PLUGIN_SMARTMIX_OPTIONS_SET'),
				showBriefly => 1,
			}]
		}) if $cb;
	});
}

sub addTracks {
	my ($client, $playFirst, $zap) = @_;

	return unless $client;
	$client = $client->master;
	return unless $prefs->client($client)->get('session_id');

	# Work out how many items need adding
	my $numItems = 0;
	my $songIndex = Slim::Player::Source::streamingSongIndex($client);
	my $songsRemaining = Slim::Player::Playlist::count($client) - $songIndex - 1;

	main::INFOLOG && $log->info("$songsRemaining songs remaining, songIndex = $songIndex");

	my $numTracks = $prefs->get('newtracks') || MIN_NEW_TRACKS;

	if ($playFirst || $zap || $songsRemaining < $numTracks) {
		my $feedback = $prefs->client($client)->get('feedback');
		
		if ($feedback && ref $feedback eq 'HASH' && keys %$feedback) {
			$feedback->{session_id} ||= $prefs->client($client->master)->get('session_id');
			
			Plugins::SmartMix::API->playlistDynamicFeedback($feedback, sub {
				my ($success) = @_;
				$prefs->client($client)->remove('feedback') if $success;
				
				_playlistDynamicNext($client, $playFirst, $zap, $songsRemaining);
			});
		}
		else {
			_playlistDynamicNext($client, $playFirst, $zap, $songsRemaining);
		}
	}
}

sub _playlistDynamicNext {
	my ($client, $playFirst, $zap, $songsRemaining) = @_;
	
	Plugins::SmartMix::API->playlistDynamicNext($prefs->client($client)->get('session_id'), NEW_TRACKS, sub {
		my $result = shift;
		
		my $songs;
		if ( $result->{response} && ($songs = $result->{response}->{songs}) && @$songs ) {
			$client->pluginData( dynamicMixFailed => 0 ) if $client->pluginData('dynamicMixFailed');
			playTracks($client, $songs, $playFirst, $zap);
		}
		else {
			my $msg = cstring($client, 'PLUGIN_SMARTMIX_NO_TRACKS');
			
			main::INFOLOG && $log->is_info && $log->error($msg . Data::Dump::dump($result));
			
			$client->showBriefly({
				'line' => [ cstring($client, 'PLUGIN_SMARTMIX'), $msg ],
				'jive' => {
					'type'    => 'popupplay',
					'text'    => [ $msg ],
				}
			}, 5) if $playFirst;
			
			# if we get an error and we're nearing the end of the list, get rid of the session_id to start a new session.
			if ( $result->{_error} && $result->{_error} =~ /400/ && $songsRemaining && (my $recipe = $prefs->client($client)->get('recipe')) ) {
				if ( !$client->pluginData('dynamicMixFailed') ) {
					main::INFOLOG && $log->info('Getting next track failed with error 400 - most likely session timed out. Let\'s try to create a new one.');
					$client->pluginData( dynamicMixFailed => 1 );
					createMix($client, undef, undef, Plugins::SmartMix::Plugin::removeCatalogIdFromRecipe($recipe), 'addToQueue');
				}
				else {
					$log->warn('Failed to get new tracks for current session. We even tried to create a new session. Giving up now.');
					# but don't stop the music! Clear session to attempt to create a new mix all together, based on current tracks etc.
					clearMixSession($client);
				}
			}
		}
	});
}

my @songQueue;
sub playTracks {
	my ($client, $songs, $playFirst, $zap, $callback) = @_;
	
	return unless $client && $songs;

	# we don't want dupes - can happen when appending a mix from a rare track
	my $playlist = { map {
		my $url = blessed($_) ? $_->url : $_;
		$url => 1;
	} @{$client->master->playlist} };
	
	my $lastPlayed = $prefs->get('noRepeatPeriod') ? time - ($prefs->get('noRepeatPeriod') * 86400) : 0;

	if ($zap && !$playFirst) {
		my $playlistCount = Slim::Player::Playlist::count($client);
		my $songIndex = Slim::Player::Source::streamingSongIndex($client) || 0;

		# Delete tracks after this one on the playlist
		for (my $i = $songIndex+1; $i < $playlistCount; $i++) {
			my $request = $client->execute(['playlist', 'delete', $songIndex+1]);
			$request->source('PLUGIN_SMARTMIX');
		}
	}
	
	@songQueue = () if $zap;
	push @songQueue, @$songs;

	my $count = 0;
		
	my $playIt = sub {
		my $item = shift || return;

		my $request = $client->execute(['playlist', ($playFirst ? 'playtracks' : 'addtracks'), 'listRef', [$item] ]);
		$request->source('PLUGIN_SMARTMIX');

		$count++ unless $playFirst;
		
		$playFirst = 0;
	};
	
	my $hasUnsupportedMusicService = Plugins::SmartMix::Services->getTrackUrl('', $client);
	
	my $work = sub {
		my ($song, $cb) = @_;
		
		if (!$song) {
			$callback->($count) if $callback;
			
			Slim::Utils::Timers::setTimer(undef, time + 5, sub {
				$client->currentPlaylistUpdateTime(Time::HiRes::time());
			});
			
			return;
		}

		my ($artist, $title, $tracks);
		if (ref $song eq 'HASH') {
			$artist = $song->{artist_name};
			$title  = $song->{title};
			
			# sometimes we have multiple IDs for one track - pick a random one...
			if ($song->{foreign_ids} && scalar @{$song->{foreign_ids}}) {
				my $songIds = $song->{foreign_ids};
				$song = $songIds->[rand @$songIds]->{foreign_id} if scalar @$songIds;
				main::DEBUGLOG && $log->is_debug && $log->debug("Found ID for local track: $song");
			}
			elsif ($song->{tracks} && scalar @{$song->{tracks}}) {
				$tracks = $song->{tracks};
#				$song = $tracks->[rand @$tracks]->{foreign_id};
				$song = $tracks->[0]->{foreign_id};
				main::DEBUGLOG && $log->is_debug && $log->debug("Found ID for online service track: $song");
			}
			elsif ($hasUnsupportedMusicService) {
				$song = $song->{id};
				main::DEBUGLOG && $log->is_debug && $log->debug("Found track ID, but no officially supported service: $song");
			}
		}
		
		# at this point $song must be a string or something went wrong
		if (ref $song) {
			main::INFOLOG && $log->info("No usable ID found in track:" . Data::Dump::dump($song));
		}
		# local tracks
		elsif ( my ($id) = $song =~ /^CA[A-Z\d]{16}:song:(.*)/ ) {
			main::DEBUGLOG && $log->is_debug && $log->debug("Trying to find local file for ID $song");
		
			my $track = Slim::Schema->rs('Track')->search({ urlmd5 => $id })->single;

			if ($track && !$playlist->{$track->url} && !_recentlyPlayed($track, $lastPlayed)) {
				$playIt->($track);
			}
		}
		# tracks from music services
		elsif ( my $track = Plugins::SmartMix::Services->getTrackUrl($song, $client, $tracks) ) {
			main::INFOLOG && $log->is_info && $log->info("Found remote tracks: " . Data::Dump::dump($track));

			my $handler = sub {
				my $track = shift;

				main::INFOLOG && $log->is_info && $log->info("Adding track to playlist: " . Data::Dump::dump($track));

				if (!$playlist->{$track}) {
					$track = Slim::Schema::RemoteTrack->updateOrCreate($track, {
						remote => 1,
						extid  => $song,
					});
					
					if ($track && !_recentlyPlayed($track, $lastPlayed)) {
						$playIt->($track);
					}
				}
			};
			
			# music services without rosetta ID need to resolve the URL asynchronously
			if (ref $track eq 'ARRAY') {
				if (main::DEBUGLOG && $log->is_debug) {
					$log->debug("No officially supported track found, trying list on unsupported services");
					$log->debug("Artist: $artist - Title: $title");
				}
				
				if ($song || ($artist && $title)) {
					my $finder = shift @$track;
					
					if (!$finder) {
						$cb->(shift @songQueue, $cb);
						return;
					}
					
					my $args = {
						artist => $artist || '',
						title  => $title ? Plugins::SmartMix::Services->cleanupTitle($title) : '',
						id     => $song,
					};
					
					my $cb2;
					
					$cb2 = sub {
						my $trackUrl = shift;
						
						if ($trackUrl) {
							$handler->($trackUrl);
							$cb->(shift @songQueue, $cb);
						}
						elsif (my $finder = shift @$track) {
							$finder->($cb2, $args);
						}
						else {
							$cb->(shift @songQueue, $cb);
						}
					};
					
					$finder->($cb2, $args);
	
					return;
				}
			} 
			else {
				$handler->($track);
			}
		}
		else {
			main::INFOLOG && $log->info("No usable ID found in track:" . Data::Dump::dump($song));
		}
		
		$cb->(shift @songQueue, $cb);
	};
	
	$work->(shift @songQueue, $work);
}

sub _recentlyPlayed {
	my ($track, $lastPlayed) = @_;
	
	my $recentlyPlayed = $lastPlayed && $track->lastplayed && $track->lastplayed > $lastPlayed;
	
	main::INFOLOG && $recentlyPlayed && $log->info("Skipping track as it has been recently played: " . $track->url . ' (' . ((time - $track->lastplayed) / 86400) . ' days ago)');
	return $recentlyPlayed;
}

sub neverStopTheMusic {
	my ($client) = @_;
	
	my $songIndex = Slim::Player::Source::streamingSongIndex($client) || 0;
	my $songsRemaining = Slim::Player::Playlist::count($client) - $songIndex - 1;

	main::INFOLOG && $log->info("$songsRemaining songs remaining, songIndex = $songIndex");

	my $numTracks = $prefs->get('newtracks') || MIN_NEW_TRACKS;
	
	if ($songsRemaining < $numTracks) {
		# grab five random artists or tracks to seed the mix (if available)
		my ($trackId, $artist, $title, $duration, %tracks, %artists);
		
		foreach (@{ Slim::Player::Playlist::playList($client) }) {
			($artist, $title, $duration) = getMixablePropertiesFromTrack($client, $_);
			next unless defined $artist && defined $title;
	
			$artists{lc($artist)}++;
			$trackId = getTrackId($_);
			$tracks{$trackId}++;
		}

		# pick five random tracks/artists from the playlist
		my $tracks = [ keys %tracks ];
		if (scalar @$tracks > 5) {
			Slim::Player::Playlist::fischer_yates_shuffle($tracks);
			splice(@$tracks, 5);
		}
		
		my $artists = [ keys %artists ];
		if (scalar @$artists > 5) {
			Slim::Player::Playlist::fischer_yates_shuffle($artists);
			splice(@$artists, 5);
		}

		# don't seed from radio stations - only do if we're playing from some track based source
		if ($tracks && $artist && $title && $duration) {
			main::INFOLOG && $log->info("Auto-mixing from random tracks in current playlist");
			
			my $artistNames = join(', ', @$artists);
			
			# Sevaral attempts to create some mix based on the last track of the current playlist:
			# 1. song mix based on five random tracks from the current playlist
			# 2. artist radio based on five random artists from the current playlist
			# 3. song mix based on the last track
			# 4. artist radio based on the last track's artist
			# 5. just some random mix based on the user's catalog
			_dontStopTheMusic($client, [{
				recipe => {
					track_id => $tracks,
					title    => $artistNames,
				},
				'warn' => "Auto-mixing from random artists in current playlist: " . join(', ', @$artists),
			},{
				recipe => {
					title  => $artistNames,
					artist => $artists,
				},
				'warn' => "Auto-mixing from last track in playlist: $title - $artist",
			},{
				recipe => {
					track_id => $trackId,
					title    => $title,
					artist   => $artist,
				},
				'warn' => "Trying an artist radio instead: $artist",
			},{
				recipe => {
					artist => $artist
				},
				error  => "...guess I have to give up being smart? I tried hard, but neither song nor artist mix worked. I'm sorry if the following mix isn't exactly a match of the previous tracks...",
			},{
				recipe => {
					type => 'catalog-radio',
				}
			}]);
		}
		elsif (main::INFOLOG && $log->is_info) {
			$log->info("No mixable items found in current playlist!");
		}
	}
}

sub _dontStopTheMusic {
	my ($client, $queue) = @_;
	
	my $next = shift @$queue;
	
	return unless $next;
	
	createMix($client, sub {
		if ( !$prefs->client($client->master)->get('session_id') ) {
			$log->error($next->{error}) if $next->{error};
			$log->warn($next->{'warn'}) if $next->{'warn'};
			
			# don't fire immediately if we were blocked due to the rate limit during the last call
			my $limit = Plugins::SmartMix::API->last_ratelimit();
			$limit->{ttl}   ||= 60;
			$limit->{limit} ||= 99;
			
			Slim::Utils::Timers::killTimers($client, \&_dontStopTheMusic);
			Slim::Utils::Timers::setTimer($client, time + ($limit->{limit} / $limit->{ttl} < 1 ? $limit->{ttl} : 0), \&_dontStopTheMusic, $queue) if scalar @$queue;
		}
	}, undef, $next->{recipe}, 'addToQueue', 'dontZap');
}

sub getMixablePropertiesFromTrack {
	my ($client, $track) = @_;
	
	return unless blessed $track;
	
	# skip tracks with a genre to be ignored
	return if Plugins::SmartMix::Importer::isIgnoredGenre($track);

	my $artist = $track->artistName;
	my $title  = $track->title;
	my $duration = $track->duration;
				
	# we might have to look up titles for remote sources
	if ( !($artist && $title && $duration) && $track && $track->remote && (my $url = $track->url) ) {
		my $handler = Slim::Player::ProtocolHandlers->handlerForURL($url);
		if ( $handler && $handler->can('getMetadataFor') ) {
			my $remoteMeta = $handler->getMetadataFor( $client, $url );
			$artist   ||= $remoteMeta->{artist};
			$title    ||= $remoteMeta->{title};
			$duration ||= $remoteMeta->{duration};
		}
	}

	return ($artist, $title, $duration);
}

sub getTrackId {
	my $track = shift;
	
	return unless blessed $track;
	
	return $track->remote ? ($track->extid || Plugins::SmartMix::Services->getTrackIdFromUrl($track->url)) : $track->urlmd5
}

sub onSkip {
	my $request = shift;
	my $client  = $request->client();
	
	return unless $prefs->client($client->master)->get('session_id');

	my $index = $request->getParam('_index');
	
	return unless $index && $index * 1 > 0;

	my $trackUrl = eval {
		if ( $client->playingSong && (my $track = $client->playingSong->track) ) {
			getTrackId($track);
		};
	};

	feedback($client, undef, undef, {
		skip_song => $trackUrl
	}) if $trackUrl;
}

1;