package Plugins::SmartMix::Plugin;

# TODO:
# - recipe sharing?
# - keep track of recent mixes so we can re-use those configurations?
# - allow to insert a static mix playlist at the current position in a queue

use strict;
use base qw(Slim::Plugin::OPMLBased);
use JSON::XS::VersionOneAndTwo;
use Scalar::Util qw(blessed);
use Storable;
use URI::Escape qw(uri_escape_utf8);

use vars qw($VERSION);

use Slim::Utils::Alarm;
use Slim::Utils::Log;
use Slim::Utils::Prefs;
use Slim::Utils::Strings qw(cstring);

use constant NOMYSB => $INC{'Slim/Networking/SqueezeNetwork.pm'} ? 1 : 0;

use Plugins::SmartMix::API;
use Plugins::SmartMix::Importer;
use Plugins::SmartMix::Services;

if (!main::SCANNER) {
	require Plugins::SmartMix::Mixer;
	
	# we use these really often...
	*createMix  = \&Plugins::SmartMix::Mixer::createMix;
	*restartMix = \&Plugins::SmartMix::Mixer::restartMix;

	if (main::WEBUI) {
		require Plugins::SmartMix::Settings;
		require Plugins::SmartMix::PlayerSettings;
	}
}

use constant THE_BLENDER_URL=> 'plugins/SmartMix/blender.html';
use constant THE_BLENDER_SRC=> 'http://www.herger.net/slim-plugins/blender/';
use constant JSONP_URL      => 'plugins/SmartMix/js.html';
use constant GRACE_PERIOD   => 45;      # inactive days after which the plugin disables itself to save catalogs
use constant FEEDBACK       => 0;		# allow user to give feedback - experimental

# SqueezePad?
my $WEBLINK_SUPPORTED_UA_RE = qr/iPeng|OrangeSqueeze/i;

my $prefs = preferences('plugin.smartmix');

my $log = Slim::Utils::Log->addLogCategory( {
	category     => 'plugin.smartmix',
	defaultLevel => 'ERROR',
	description  => 'PLUGIN_SMARTMIX',
} );

our $recipes = [{
	name => 'Smooth',
	type => 'artist-description',
	mood => 'calming',
	max_energy => 0.4,
	max_tempo  => 80,
},{
	name   => "Let's Dance!",
	type   => 'artist-description',
	'sort' => 'tempo-asc',
	description => 'disco',
	min_danceability => 0.7,
},{
	name  => 'Play it Loud!',
	type  => 'artist-description',
	style => 'metal',
	min_energy => 0.7,
	min_tempo  => 120,
},{
	name  => 'Classic Rock',
	type  => 'artist-description',
	style => ['classic rock^2','rock','-r&b'],
}];

my %feedbackCache;

sub initPlugin {
	my $class = shift;

	$VERSION = $class->_pluginDataFor('version');
	Plugins::SmartMix::API->init($class->_pluginDataFor('aid'), $VERSION);
	
	# We only have a very limit number of catalogs available (1000). In order to save these resources
	# we're going to automatically delete catalogs which have not been used for more than GRACE_PERIOD days.
	if (!$prefs->get('dontDisable') && $prefs->get('lastused') && ($prefs->get('lastused') + 86400 * GRACE_PERIOD) < time) {
		$log->error("SmartMix has not been used for more than " . GRACE_PERIOD . " days - disabling plugin to save resources.");
			
		my $songCatalogId = $prefs->get('songCatalogId');
		Plugins::SmartMix::API->catalogDelete($songCatalogId, sub {
			$log->error("SmartMix catalog ID $songCatalogId has been deleted. - Thanks for your understanding.");
		}) if $songCatalogId;
		
		Slim::Utils::PluginManager->disablePlugin('SmartMix');
		$prefs->set('lastused', 0);
		
		return;
	}

	if (!main::SCANNER) {
		$prefs->setChange( sub {
			main::INFOLOG && $log->info("Song catalog ID changed: $_[3] -> $_[1] - trigger rescan of the music collection");

			$prefs->set('wipe', 1);

			Plugins::SmartMix::API->catalogUpdate($_[1], [{
				action => 'update',
				catalog_keyvalues => {
					# TEN sometimes complains about invalid JSON - make sure we treat time as a string?
					created => time . '',
					version => $VERSION
				}
			}], 1);

			$prefs->set('lastused', time) if ($prefs->get('lastused') || 0) + 86400 < time;

			# try not to immediately trigger the scan - only available on 7.9+
			eval { Slim::Music::Import->doQueueScanTasks(1); };
			Slim::Control::Request::executeRequest( undef, [ 'rescan' ] );
			eval { Slim::Music::Import->doQueueScanTasks(0); };
		}, 'songCatalogId');

		$prefs->setChange( sub {
			my $client = $_[2] || return;
			my $newValue = $_[1];
			restartMix($client);
		}, 'localOnly');

		# read country setting from Triode's spotify plugin
		my $spotifyPrefs = preferences('plugin.spotify');
		$spotifyPrefs->setChange( sub {
			$prefs->set('spotifyCountry', $_[1]) if $_[1];
		}, 'location') if $spotifyPrefs;

		# a few default recipes and prefs
		$prefs->init({
			recipes => $recipes,
			noRepeatPeriod => 0,
			spotifyCountry => $spotifyPrefs ? sub {
				return $spotifyPrefs->get('location') || undef;
			} : undef,
		});

		# migrate neverStopTheMusic from global to per-player; use old value as default for new players
		$prefs->migrateClient(2, sub {
			my ($cprefs, $client) = @_;
			$cprefs->set('neverStopTheMusic', $prefs->get('neverStopTheMusic'));
			1;
		});
		
		$prefs->setValidate({ 'validator' => 'intlimit', 'low' => 0, 'high'=> 180 }, 'noRepeatPeriod');
	}
	
	Plugins::SmartMix::Importer->initPlugin();

	$class->initSongCatalogId();

	if (!main::SCANNER) {
		require Tie::Cache::LRU::Expires;
		tie my %feedbackCache, 'Tie::Cache::LRU::Expires', EXPIRES => 45, ENTRIES => 10;

		Plugins::SmartMix::Mixer->init();

		Slim::Control::Request::addDispatch(['smartmix', 'ismixing'], [1, 1, 0, \&cliIsMixing]);
		Slim::Control::Request::addDispatch(['smartmix', 'like'], [1, 0, 0, \&cliFeedback]);
		Slim::Control::Request::addDispatch(['smartmix', 'dislike'], [1, 0, 0, \&cliFeedback]);
		Slim::Control::Request::addDispatch(['smartmix', 'blenderlink'], [1, 1, 0, \&cliBlenderLink]);

		Slim::Control::Request::subscribe(sub {
			$prefs->remove('wipe');
		}, [['rescan'], ['done']] );
	
		# Track Info item while mixing
		Slim::Menu::TrackInfo->registerInfoProvider( smartmix_playing => (
			isa => 'top',
			func   => \&nowPlayingInfoMenu,
		) );
		
		# Track Info item
		Slim::Menu::TrackInfo->registerInfoProvider( smartmix_songmix => (
			before => 'middle',
			func   => \&trackInfoMenu,
		) );
	
		# Artist Info item
		Slim::Menu::ArtistInfo->registerInfoProvider( smartmix_artistmix => (
			after => 'middle',
			func  => \&artistInfoMenu,
		) );
	
		# Album Info item
		Slim::Menu::AlbumInfo->registerInfoProvider( smartmix_albummix => (
			after => 'middle',
			func  => \&albumInfoMenu,
		) );
	
		if (main::WEBUI) {
			# very simple page handler to load "The Blender" in a separate window
			Slim::Web::Pages->addPageFunction(THE_BLENDER_URL, \&handleWeb);
		
			Plugins::SmartMix::Settings->new();
			Plugins::SmartMix::PlayerSettings->new();
		}
		
		# I'm calling this js.html as our HTTP code is setting the right content type for this file...
		Slim::Web::Pages->addPageFunction(JSONP_URL, \&handleJSONP);
		
		require Plugins::SmartMix::ProtocolHandler;
		Slim::Player::ProtocolHandlers->registerHandler(
			smartmix => 'Plugins::SmartMix::ProtocolHandler'
		);
		
		# Update whenever we change recipes
		_initAlarms();
		$prefs->setChange( \&_initAlarms, 'recipes' );
	}
	
	$class->SUPER::initPlugin(
		feed   => \&handleFeed,
		tag    => 'smartmix',
		node   => 'myMusic',
		menu   => 'browse',
		weight => 80,
	);
}

sub getDisplayName { 'PLUGIN_SMARTMIX' }

sub handleFeed {
	my ($client, $cb, $args) = @_;
	
	# try to get user's country setting from mysb.com
	if ( !$prefs->get('spotifyCountry') && !NOMYSB ) {
		Slim::Networking::SqueezeNetwork->new(
			sub {
				eval {
					my $profile = from_json($_[0]->content);
					$prefs->set('spotifyCountry', $profile->{country});
				};
			},
			sub {},
			{ client => $client },
		)->get(	Slim::Networking::SqueezeNetwork->url('/api/spotify/v1/opml/profile') );
	}
	
	if ( !$prefs->get('songCatalogId') ) {
		$cb->({
			items => [{
				name => cstring($client, 'PLUGIN_SMARTMIX_NO_CATALOG'),
				type => 'textarea',
			}],
		});
		return;
	}
	
	my $items = [{
		name => cstring($client, 'PLUGIN_SMARTMIX_OPTIONS'),
		url  => \&mixerOptionsMenu,
	},{
		name => cstring($client, 'PLUGIN_SMARTMIX_START_FROM_CURRENT_TRACK'),
		url  => \&Plugins::SmartMix::Mixer::createMixFromCurrentTrack,
		type => 'playlist',
		nextWindow => 'nowPlaying'
	}];

	my $totals = $prefs->get('ignore_countWarning') ? 0 : Slim::Schema->totals;
	if ( $totals && $totals->{track} > 100_000 ) {
		my $info = cstring($client, 'PLUGIN_SMARTMIX_TOO_MANY_TRACKS_DESC');
		$info =~ s/<.*?>//g;
		
		unshift @$items, {
			name => cstring($client, 'PLUGIN_SMARTMIX_TOO_MANY_TRACKS'),
			items => [{
				name => $info . "\n\n" . cstring($client, 'PLUGIN_SMARTMIX_TOO_MANY_TRACKS_DISABLE_WARNING'),
				type => 'textarea',
			}],
		};
	}

	# if user has a music service, provide a "Recommendation" mix
	if ( Plugins::SmartMix::Services->getServices($client) ) {
		push @$items, {
			name => cstring($client, 'PLUGIN_SMARTMIX_RECOMMENDED'),
			url  => 'smartmix://_recommended',
			type => 'audio',
			nextWindow  => 'nowPlaying',
		};
	}
	
	# mix based on the listeners most listened to artists
	if ( main::STATISTICS ) {
		push @$items, {
			name => cstring($client, 'PLUGIN_SMARTMIX_TOP_ARTISTS'),
			url  => 'smartmix://_top_artists',
			type => 'audio',
			nextWindow  => 'nowPlaying',
		};
	}
	
	# read our recipes from the prefs
	foreach my $recipe ( @{ getRecipes() } ) {
		push @$items, {
			name => $recipe->{name},
			url  => 'smartmix://' . uri_escape_utf8(lc($recipe->{name})),
			type => 'audio',
			passthrough => [ $recipe ],
			nextWindow  => 'nowPlaying',
		},
	}
	
	push @$items, {
		name => cstring($client, 'PLUGIN_SMARTMIX_THE_BLENDER'),
		weblink => '/' . THE_BLENDER_URL,
	} if $args->{isWeb};
	
	push @$items, {
		name => cstring($client, 'PLUGIN_SMARTMIX_THE_BLENDER'),
		itemActions => {
			items => {
				command  => [ 'smartmix', 'blenderlink' ],
			},
		},
	} if __PACKAGE__->canWeblink($client);
	
	$cb->({
		items => $items
	});
}

sub mixerOptionsMenu {
	my ($client, $cb, $args, $isContextMenu) = @_;
	
	my $items = [];
	
	my $type = 'playlist' if $args->{isButton};
	
	# add item to controll the currently playing mix
	if ( $prefs->client($client->master)->get('session_id') ) {
		my $trackUrl = eval {
			if ( $client->playingSong && (my $track = $client->playingSong->track) ) {
				Plugins::SmartMix::Mixer::getTrackId($track);
			};
		};
		
		push @$items, {
			name => cstring($client, 'PLUGIN_SMARTMIX_MORE_LIKE_THIS'),
			url  => \&Plugins::SmartMix::Mixer::steer,
			type => $type,
			passthrough => [{
				more_like_this => $trackUrl
			}],
			nextWindow => $args->{isButton} ? 'parent' : 'refresh',
		},{
			name => cstring($client, 'PLUGIN_SMARTMIX_LESS_LIKE_THIS'),
			url  => \&Plugins::SmartMix::Mixer::steer,
			type => $type,
			passthrough => [{
				less_like_this => $trackUrl
			}],
			nextWindow => $args->{isButton} ? 'parent' : 'refresh',
		} if $trackUrl;
		
		push @$items, {
			name => cstring($client, 'PLUGIN_SMARTMIX_BAN_ARTIST'),
			url  => \&Plugins::SmartMix::Mixer::feedback,
			type => $type,
			passthrough => [{
				ban_artist => $trackUrl
			}],
			nextWindow => $args->{isButton} ? 'parent' : 'refresh',
		} if FEEDBACK && $trackUrl;
		
		push @$items, {
			name => cstring($client, 'PLUGIN_SMARTMIX_HOTTTNESSS'),
			type => 'outline',
			items => [{
				name => cstring($client, 'PLUGIN_SMARTMIX_HOTTTNESSS_VERRRY'),
				url => \&Plugins::SmartMix::Mixer::steer,
				type => $type,
				passthrough => [{
					min_artist_hotttnesss  => 0.6,
					max_artist_hotttnesss  => 1,
					max_artist_familiarity => 1,
					min_song_hotttnesss    => 0.5,
					max_song_hotttnesss    => 1,
				}],
				nextWindow => 'parent',
			},{
				name => cstring($client, 'PLUGIN_SMARTMIX_HOTTTNESSS_DEFAULT'),
				url => \&Plugins::SmartMix::Mixer::steer,
				type => $type,
				passthrough => [{
					min_artist_hotttnesss  => 0,
					max_artist_hotttnesss  => 1,
					max_artist_familiarity => 1,
					min_song_hotttnesss    => 0,
					max_song_hotttnesss    => 1,
				}],
				nextWindow => 'parent',
			},{
				name => cstring($client, 'PLUGIN_SMARTMIX_HOTTTNESSS_NOT'),
				url => \&Plugins::SmartMix::Mixer::steer,
				type => $type,
				passthrough => [{
					min_artist_hotttnesss  => 0,
					max_artist_hotttnesss  => 0.4,
					max_artist_familiarity => 0.8,
					min_song_hotttnesss    => 0,
					max_song_hotttnesss    => 0.6,
				}],
				nextWindow => 'parent',
			}],
			isContextMenu => $isContextMenu,
		};
		
		if ( _mixHasArtist($client) ) {
			push @$items, {
				name => cstring($client, 'PLUGIN_SMARTMIX_VARIETY'),
				type => 'outline',
				items => [{
					name => cstring($client, 'PLUGIN_SMARTMIX_ARTIST_ONLY'),
					url => \&restartMix,
					type => $type,
					passthrough => [{
						variety => 0,
						type => 'artist'
					}],
					nextWindow => 'parent',
				},{
					name => cstring($client, 'PLUGIN_SMARTMIX_VARIETY_LITTLE'),
					url => \&restartMix,
					type => $type,
					passthrough => [{
						variety => 0.1,
						type => 'artist-radio'
					}],
					nextWindow => 'parent',
				},{
					name => cstring($client, 'PLUGIN_SMARTMIX_VARIETY_MEDIUM'),
					url => \&restartMix,
					type => $type,
					passthrough => [{
						variety => 0.5,
						type => 'artist-radio'
					}],
					nextWindow => 'parent',
				},{
					name => cstring($client, 'PLUGIN_SMARTMIX_VARIETY_LARGE'),
					url => \&restartMix,
					type => $type,
					passthrough => [{
						variety => 0.9,
						type => 'artist-radio'
					}],
					nextWindow => 'parent',
				}],
				isContextMenu => $isContextMenu,
			};
		}
		elsif ( _isCatalogRadio($client) ) {
			push @$items, {
				name => cstring($client, 'PLUGIN_SMARTMIX_VARIETY'),
				type => 'outline',
				items => [{
					name => cstring($client, 'PLUGIN_SMARTMIX_ADVENTUROUSNESS_HIGH'),
					url => \&Plugins::SmartMix::Mixer::steer,
					type => $type,
					passthrough => [{
						adventurousness => 1
					}],
					nextWindow => 'parent',
				},{
					name => cstring($client, 'PLUGIN_SMARTMIX_ADVENTUROUSNESS_MEDIUM'),
					url => \&Plugins::SmartMix::Mixer::steer,
					type => $type,
					passthrough => [{
						adventurousness => 0.6,
					}],
					nextWindow => 'parent',
				},{
					name => cstring($client, 'PLUGIN_SMARTMIX_VARIETY_MEDIUM'),
					url => \&Plugins::SmartMix::Mixer::steer,
					type => $type,
					passthrough => [{
						adventurousness => 0.2,
					}],
					nextWindow => 'parent',
				},{
					name => cstring($client, 'PLUGIN_SMARTMIX_ADVENTUROUSNESS_LOW'),
					url => \&Plugins::SmartMix::Mixer::steer,
					type => $type,
					passthrough => [{
						adventurousness => 0.05,
					}],
					nextWindow => 'parent',
				}],
				isContextMenu => $isContextMenu,
			};
		}
		
		push @$items, {
			name => cstring($client, 'PLUGIN_SMARTMIX_ENERGY'),
			type => 'outline',
			items => [{
				name => cstring($client, 'PLUGIN_SMARTMIX_MORE_ENERGY'),
				url => \&Plugins::SmartMix::Mixer::steer,
				type => $type,
				passthrough => [{
					max_energy => 1,
					min_energy => 0.3,
					max_tempo => 500,
					min_tempo => 80
				}],
				nextWindow => 'parent',
			},{
				name => cstring($client, 'PLUGIN_SMARTMIX_VARIETY_MEDIUM'),
				url => \&Plugins::SmartMix::Mixer::steer,
				type => $type,
				passthrough => [{
					max_energy => 1,
					min_energy => 0,
					max_tempo => 500,
					min_tempo => 0
				}],
				nextWindow => 'parent',
			},{
				name => cstring($client, 'PLUGIN_SMARTMIX_LESS_ENERGY'),
				url => \&Plugins::SmartMix::Mixer::steer,
				type => $type,
				passthrough => [{
					max_energy => 0.6,
					min_energy => 0,
					max_tempo => 130,
					min_tempo => 0
				}],
				nextWindow => 'parent',
			}],
			isContextMenu => $isContextMenu,
		},{
			name => cstring($client, 'PLUGIN_SMARTMIX_DANCEABILITY'),
			type => 'outline',
			items => [{
				name => cstring($client, 'PLUGIN_SMARTMIX_DANCEABILITY_HIGH'),
				url => \&Plugins::SmartMix::Mixer::steer,
				type => $type,
				passthrough => [{
					max_danceability => 1,
					min_danceability => 0.5,
				}],
				nextWindow => 'parent',
			},{
				name => cstring($client, 'PLUGIN_SMARTMIX_VARIETY_MEDIUM'),
				url => \&Plugins::SmartMix::Mixer::steer,
				type => $type,
				passthrough => [{
					max_danceability => 1,
					min_danceability => 0,
				}],
				nextWindow => 'parent',
			},{
				name => cstring($client, 'PLUGIN_SMARTMIX_LESS_ENERGY'),
				url => \&Plugins::SmartMix::Mixer::steer,
				type => $type,
				passthrough => [{
					max_danceability => 0.6,
					min_danceability => 0,
				}],
				nextWindow => 'parent',
			}],
			isContextMenu => $isContextMenu,
		};
		
		push @$items, {
			name => cstring($client, "PLUGIN_SMARTMIX_SAVE_RECIPE"),
			type => 'search',
			url  => \&addRecipe,
			nextWindow => 'parent',
			isContextMenu => $isContextMenu,
		}
	}
	elsif ( $args->{isButton} ) {
		$items = [{
			name => cstring($client, 'PLUGIN_SMARTMIX_NO_MIX_PLAYING'),
			type => 'text',
		}];
	}
		
	if ( Plugins::SmartMix::Services->getServices($client) ) {
		push @$items, {
			name => cstring($client, "PLUGIN_SMARTMIX_SOURCES"),
			type => 'outline',
			items => [{
				name => cstring($client, 'PLUGIN_SMARTMIX_LOCAL_MUSIC'),
				url => sub {
					$prefs->client($_[0])->set('localOnly', 1 );
					restartMix(@_) if $prefs->client($_[0])->get('session_id');
				},
				type => $type,
				nextWindow => 'parent',
			},{
				name => cstring($client, 'PLUGIN_SMARTMIX_ANY_SOURCE'),
				url => sub {
					$prefs->client($_[0])->set('localOnly', 0 );
					restartMix(@_) if $prefs->client($_[0])->get('session_id');
				},
				type => $type,
				nextWindow => 'parent',
			}],
			isContextMenu => $isContextMenu,
		};
	}

	if ( !$args->{isButton} ) {
		push @$items, {
			name => cstring($client, 'PLUGIN_SMARTMIX_DONT_STOP_THE_MUSIC') . cstring($client, 'COLON') . ' ' . cstring($client, $prefs->client($client)->get('neverStopTheMusic') ? 'ON' : 'OFF'),
			url => sub {
				my $cb = $_[1];
				$prefs->client($client)->set('neverStopTheMusic', $prefs->client($client)->get('neverStopTheMusic') ? 0 : 1);
				$cb->() if $cb;
			},
			type => $type,
			nextWindow => 'refresh',
		};
	}
	
	$cb->({
		items => $items
	});
}

sub _mixHasArtist {
	my $client = shift;
	
	my $recipe = $prefs->client($client->master)->get('recipe') || return;
	return $recipe->{artist} || $recipe->{artist_id} || $recipe->{song_id} || $recipe->{track_id};
}

sub _isCatalogRadio {
	my $client = shift;
	
	my $recipe = $prefs->client($client->master)->get('recipe') || return;
	return $recipe->{seed_catalog} && $recipe->{type} =~ /catalog/i;
}

sub trackInfoMenu {
	my ( $client, $url, $track, $remoteMeta, $tags ) = @_;

	return unless $prefs->get('songCatalogId');

	my $artist   = $track->remote ? $remoteMeta->{artist}   : $track->artistName;
	my $title    = $track->remote ? $remoteMeta->{title}    : $track->title;
	my $duration = $track->remote ? $remoteMeta->{duration} : $track->duration;

	my $items = _artistInfoMenu($client, $artist, $duration);

	unshift @$items, {
		name => cstring($client, 'PLUGIN_SMARTMIX_MIX', $title),
		url  => \&createMix,
		type => 'playlist',
		passthrough => [{
			track_id => Plugins::SmartMix::Mixer::getTrackId($track),
			title    => $title,
			artist   => $artist,
		}, $duration],
		nextWindow => 'parent',
	} if $title;
	
	return $items;
}

sub nowPlayingInfoMenu {
	my ( $client, $url, $track, $remoteMeta, $tags ) = @_;

	return unless $prefs->get('songCatalogId') && $client;
	
	my $items = [];

	# add item to control the current playlist
	if ( $prefs->client($client->master)->get('session_id') && $client->playingSong && $client->playingSong->track->id == $track->id ) {
		$items = [{
			name => cstring($client, 'PLUGIN_SMARTMIX_OPTIONS'),
			url  => \&mixerOptionsMenu,
			type => 'outline',
			passthrough => [ 1 ],
			isContextMenu => 1,
		}];
	}
	
	return $items;
}

sub artistInfoMenu {
	my ( $client, $url, $artist, $remoteMeta, $tags ) = @_;

	return unless $prefs->get('songCatalogId');
	
	return _artistInfoMenu( $client, $artist->name || $remoteMeta->{artist} );
}

sub albumInfoMenu {
	my ( $client, $url, $album, $remoteMeta, $tags ) = @_;

	return unless $prefs->get('songCatalogId');
	
	return _artistInfoMenu( $client, $album->contributor->name || $remoteMeta->{artist} );
}

sub _artistInfoMenu {
	my ( $client, $artist, $duration ) = @_;

	my $items = [];
	
	push @$items, {
		name => cstring($client, 'PLUGIN_SMARTMIX_MIX', $artist),
		url => \&createMix,
		type => 'playlist',
		passthrough => [{
			artist => $artist,
		}, $duration],
		nextWindow => 'parent',
	} if $artist;

	return $items;
}
sub addRecipe {
	my ($client, $cb, $params) = @_;

	if ($params->{search} && (my $args = $prefs->client($client->master)->get('recipe')) ) {
		$args->{name} = $params->{search};
		saveRecipe({
			args => $args,
		});
	}

	$cb->({
		items => [{
			name => cstring($client, 'PLUGIN_SMARTMIX_RECIPE_SAVED', $params->{search}),
			showBriefly => 1,
		}]
	});
}

sub saveRecipe {
	my ($args) = @_;
	
	return unless $args && $args->{args} && $args->{args}->{name};
	
	my $name = $args->{args}->{name};
	
	my @recipes = grep { lc($_->{name}) ne lc($name) } @{ getRecipes() };
	push @recipes, removeCatalogIdFromRecipe($args->{args});

	$prefs->set('recipes', \@recipes);	
}

sub removeCatalogIdFromRecipe {
	my $args = shift;
	
	return $args unless $args && $args->{bucket};
	
	@{$args->{bucket}} = grep { $_ !~ /^(?:id:|track)/ } @{$args->{bucket}};
	delete $args->{bucket} unless scalar @{$args->{bucket}};
	delete $args->{limit};
	
	return $args;
}

sub initSongCatalogId {
	my $class = shift;
	
	Plugins::SmartMix::API->catalogProfile( preferences('server')->get('server_uuid'), 'song', sub {
		my $result = shift;
		
		if ( $result->{response} && $result->{response}->{catalog} && (my $id = $result->{response}->{catalog}->{id}) ) {
			$prefs->set('songCatalogId', $id);
		}
		else {
			$class->createCatalogId();
		}
	} );
}

sub createCatalogId {
	my $uuid = preferences('server')->get('server_uuid');

	Plugins::SmartMix::API->catalogCreate( $uuid, 'song', sub {
		my $result = shift;
		
		if ( $result->{response} && $result->{response}->{id} ) {
			main::INFOLOG && $log->is_info && $log->info("Got new catalog ID: " . $result->{response}->{id});
			$prefs->set('songCatalogId', $result->{response}->{id});
		}
		else {
			# if we've even failed async creation, then try it the hard way, using syncronous http requests
			# while these are blocking, at least we get the full content back
			$log->warn('Failed to get catalog ID using async calls - trying using blocking calls next...');

			$result = Plugins::SmartMix::API->catalogProfile($uuid, 'song');
			my $id;
			if ( $result->{response} && $result->{response}->{catalog} && ($id = $result->{response}->{catalog}->{id}) ) {
				$prefs->set('songCatalogId', $id);
			}
			elsif ( $id = Plugins::SmartMix::API->catalogCreate($uuid, 'song') ) {
				$prefs->set('songCatalogId', $id);
			}
			else {
				$log->error('Giving up getting a catalog ID');
			}
		}
	} );
}

sub handleWeb {
	my ($client, $params, $callback, $httpClient, $httpResponse, $request) = @_;
	
	if (my $recipe = delete $params->{recipe}) {
		my ($r) = grep { lc($_->{name}) eq lc($recipe) } @{ getRecipes() };
		$params->{recipe} = to_json($r) if $r;
	}
	
	$params->{aid} = Plugins::SmartMix::API::aid();
	$params->{catalog}  = $prefs->get('songCatalogId');
	$params->{catalogs} = join( ',', @{Plugins::SmartMix::Mixer::getBuckets($client)} );
	$params->{blender_source} = THE_BLENDER_SRC;
	
	$params->{touch} ||= $params->{userAgent} && $params->{userAgent} =~ /(?:Mobile(\/|\s)|Android)/;
	
	# use $params->{host} to be sure we're using the same cookies as the main page (eg. for the player selection)
	my $host = $params->{host} || (Slim::Utils::Network::serverAddr() . ':' . preferences('server')->get('httpport'));
	$params->{squeezebox_server} = 'http://' . $host . '/' . JSONP_URL . '?player=' . uri_escape_utf8($client->id);
	
	return Slim::Web::HTTP::filltemplatefile($params->{'path'}, $params);
}

sub handleJSONP {
	my ($client, $params, $callback, $httpClient, $httpResponse, $request) = @_;

	my $response = {
		error => 'invalid arguments'
	};

	my $args;
	eval {
		$args = from_json($params->{args});
	};

	if (main::INFOLOG && $log->is_info) {
		$log->info( "Got JSONP request: " . ($@ ? $params->{args} : Data::Dump::dump($args)) );
	}

	if ($@) {
		$log->error("Failed to decode JSONP request: $@");
	}

	if ($args->{cmd} && $args->{cmd} =~ /^(?:savemix)$/ && $args->{args} && ref $args->{args} eq 'HASH' && $args->{args}->{name}) {
		my $name = $args->{args}->{name};

		$response = {
			error => 0,
			msg => "Saved mix as $name",
			args => $args->{args}
		};
		
		saveRecipe($args);
	}
	
	elsif ($client && $args->{cmd} && $args->{cmd} =~ /^(?:playtracks|createmix)$/ && ($args->{tracks} || $args->{args})) {
		Plugins::SmartMix::Mixer::clearMixSession($client);
		
		Plugins::SmartMix::Mixer::playTracks($client, $args->{tracks}, 1, undef, sub {
			my $count = shift;

			$response = {
				error => $count ? 0 : 'No playable tracks found',
				msg => "Playing $count tracks to " . ($client ? $client->name : 'nada') . '.'
			};
			
			# configure playlist to update dynamically once the static tracks have finished playing (if there are at all)
			if ($args->{args} && ref $args->{args} eq 'HASH') {
				$response ||= {};
				$response->{msg} .= " Starting mix on " . ($client ? $client->name : 'nada');
				$response->{args} = $args->{args};
				
				createMix($client, undef, undef, removeCatalogIdFromRecipe($args->{args}), $count ? 'addToQueue' : undef, $count ? 'dontZap' : undef);
			}

			my $content = $params->{callback} . '(' . to_json($response) . ')';
			$httpResponse->header( 'Content-Length' => length($content) );

			$callback->($client, $params, \$content, $httpClient, $httpResponse);
		});
		
		return;
	}

 	elsif ($client && $args->{cmd} && $args->{cmd} =~ /^(?:getsessionid)$/ ) {
 		if (! $prefs->client($client->master)->get('session_id')) {
 			$response = {
 				error => 1,
 				msg => "No session found"
 			};
 		} else {
 			$response = {
 				error => 0,
 				msg => "SID ok",
 				session_id => $prefs->client($client->master)->get('session_id')
 			};
 		}
 	}
	
	else {
		my $msg = "No client connected or invalid arguments: " . Data::Dump::dump($args);
		$log->warn($msg);
		$response = {
			error => 1,
			msg   => $msg,
		}
	}

	my $content = $params->{callback} . '(' . to_json($response) . ')';

	$httpResponse->header( 'Content-Length' => length($content) );
	
	return \$content;
}

sub _initAlarms {
	Slim::Utils::Alarm->addPlaylists('PLUGIN_SMARTMIX', [ map { {
		title => $_->{name},
		url   => 'smartmix://' . uri_escape_utf8(lc($_->{name})),
	} } @{ getRecipes() } ]);
}

sub getRecipes {
	my $recipes = [];

	foreach ( sort { lc($a->{name}) cmp lc($b->{name}) } @{ $prefs->get('recipes') } ) {
		push @$recipes, Storable::dclone($_);
	}
	
	return $recipes;
}

sub canWeblink {
	my ($class, $client) = @_;
	return $client && $client->controllerUA && $client->controllerUA =~ $WEBLINK_SUPPORTED_UA_RE;
}

sub cliBlenderLink {
	my $request = shift;
	
	if ($request->isNotQuery([['smartmix'], ['blenderlink']])) {
		$request->setStatusBadDispatch();
		return;
	}
	
	$request->setStatusProcessing();

	my $client = $request->client;

	my $i = 0;
	$request->addResultLoop('item_loop', $i, 'text', cstring($client, 'PLUGIN_SMARTMIX_THE_BLENDER_NOTE'));
	$request->addResultLoop('item_loop', $i, 'weblink', 'http://' . Slim::Utils::Network::serverAddr() . ':' . preferences('server')->get('httpport') . '/' . THE_BLENDER_URL . '?touch=1&player=' . uri_escape_utf8($client->id));

	$request->addResult('count', $i);
	$request->addResult('offset', 0);

	$request->setStatusDone();
}

sub cliIsMixing {
	my $request = shift;
	
	if ($request->isNotQuery([['smartmix'], ['ismixing']])) {
		$request->setStatusBadDispatch();
		return;
	}
	
	$request->setStatusProcessing();

	$request->addResult('_isMixing', $prefs->client( $request->client->master )->get('session_id') ? 1 : 0);

	$request->setStatusDone();
}

sub cliFeedback {
	my $request = shift;
	
	if ($request->isNotCommand([['smartmix'], ['like', 'dislike']])) {
		$request->setStatusBadDispatch();
		return;
	}

	my $client   = $request->client->master;
	my $feedback = lc($request->getRequest(1) || '');
	
	if ( $client && $feedback && $prefs->client($client->master)->get('session_id') ) {
		my $trackUrl = eval {
			if ( $client->playingSong && (my $track = $client->playingSong->track) ) {
				Plugins::SmartMix::Mixer::getTrackId($track);
			};
		};

		# use a small cache of feedback commands so we don't repeatedly send the same
		if ($trackUrl && !$feedbackCache{"$feedback:$trackUrl"}) {
			$feedbackCache{"$feedback:$trackUrl"}++;
			
			my $args = {};
			
			if ( $feedback eq 'like' ) {
				$args->{more_like_this} = $trackUrl;
				delete $feedbackCache{"dislike:$trackUrl"};
			}
			elsif ( $feedback eq 'dislike' ) {
				$args->{less_like_this} = $trackUrl;
				delete $feedbackCache{"like:$trackUrl"};
			}
			
			if (keys %$args) {
				Plugins::SmartMix::Mixer::steer($client, undef, undef, $args);
			}
		}
	}

	$request->setStatusDone();
}


1;
