package Plugins::SmartMix::Settings;

use strict;
use base qw(Slim::Web::Settings);

use File::Spec::Functions;
use File::ReadBackwards;

use Slim::Utils::Log;
use Slim::Utils::Prefs;

use Plugins::SmartMix::Plugin;
use Plugins::SmartMix::Importer;
use Plugins::SmartMix::Services;

use constant LOG_URL => 'plugins/SmartMix/smartmix-missing.log';

my $prefs = preferences('plugin.smartmix');
my $log   = logger('plugin.smartmix');
my $missing_log = catfile(Slim::Utils::OSDetect::dirsFor('log'), 'smartmix-missing.log');

my $hasVirtualLibraries;

eval {
	require Slim::Music::VirtualLibraries;
	$hasVirtualLibraries = 1;
};

sub new {
	my $class = shift;

	Slim::Web::Pages->addPageFunction(LOG_URL, \&handleLog);
	
	return $class->SUPER::new(@_);
}

sub name {
	return 'PLUGIN_SMARTMIX';
}

sub prefs {
	 my @prefs = ('ignoreGenres', 'noRepeatPeriod', 'ignore_countWarning', 'spotifyCountry', 'failsafeUpload');
	 push @prefs, 'useLibrary' if $hasVirtualLibraries;
	 return ($prefs, @prefs);
}

sub page {
	return 'plugins/SmartMix/settings.html';
}

sub handler {
	my ($class, $client, $paramRef, $pageSetup) = @_;
	
	# make sure the country setting is always uppercase
	$prefs->set('spotifyCountry', uc($prefs->get('spotifyCountry')));

	$paramRef->{services} = Plugins::SmartMix::Services->getAllServices();

	my $recipes = Plugins::SmartMix::Plugin->getRecipes();

	if ($paramRef->{cleanup}) {
		Plugins::SmartMix::Importer->cleanupCatalog();
	}
	elsif ($paramRef->{'saveSettings'}) {
		my $defaults;
		
		# remove items from the recipes list
		if ($paramRef->{'resetRecipes'}) {
			foreach (@$Plugins::SmartMix::Plugin::recipes) {
				$paramRef->{'delete_' . $_->{name}} = 1;
			}
		}
		
		my %delete;
		foreach (keys %{$paramRef}) {
			next unless /^delete_(.*)/;
			main::INFOLOG && $log->is_info && $log->info("Deleting recipe '$1'");
			$delete{$1}++
		}

		@$recipes = grep { !$delete{$_->{name}} } @$recipes;
		
		if ($paramRef->{'resetRecipes'}) {
			main::INFOLOG && $log->info('Resetting default recipes');
			unshift @$recipes, @{$Plugins::SmartMix::Plugin::recipes};
		}

		$prefs->set('recipes', $recipes);
		
		# handle music sources manually, as we want inverted logic (don't set param if entry is checked)
		foreach (@{$paramRef->{services}}, 'local') {
			my $pref = 'disable_' . $_;
			if ($paramRef->{$pref}) {
				$prefs->remove($pref);
			}
			else {
				$prefs->set($pref, 1);
			}
		}

		if ( ($paramRef->{'pref_useLibrary'} || '') ne ($prefs->get('useLibrary') || '') ) {
			$paramRef->{'resetCatalog'} = 1;
		}

		if ($paramRef->{'resetCatalog'} && (my $id = $prefs->get('songCatalogId'))) {
			main::INFOLOG && $log->info("Delete catalog $id to create a new one");
			Plugins::SmartMix::API->catalogDelete($id, sub {
				Plugins::SmartMix::Plugin->createCatalogId()
			});
		}

		Plugins::SmartMix::Services->init();
	}

	my $totals = Slim::Schema->totals;
	if ( $totals->{track} > 100_000 ) {
		$paramRef->{countWarning} = 1;
	}

	eval {
		my %locations = ('' => '');
		for my $code (Locale::Country::all_country_codes()) {
			$locations{Locale::Country::code2country($code)} = uc $code;
		}
		$paramRef->{'locations'} = \%locations;
	};

# we're now falling back to looking up tracks over a proxy - no need for SSL any more
#	$paramRef->{noSSL} = !Slim::Networking::Async::HTTP::hasSSL();
	
	$paramRef->{prefs} = $prefs->all();
	$paramRef->{recipes} = $recipes;

	$paramRef->{cleanupRunning} = Plugins::SmartMix::Importer->isCleanupRunning();
	
	if ( $hasVirtualLibraries ) {
		my $libraries = Slim::Music::VirtualLibraries->getLibraries();
		%{ $paramRef->{libraries} } = map {
			my $count = Slim::Utils::Misc::delimitThousands(Slim::Music::VirtualLibraries->getTrackCount($_));
			$_ => $libraries->{$_}->{name} . sprintf(" (%s %s)", $count, Slim::Utils::Strings::cstring($client, 'SONGS'));
		} keys %$libraries if keys %$libraries;
	}

	$paramRef->{missing_log} = LOG_URL;
	$paramRef->{missing_log_path} = $missing_log;
	
	return $class->SUPER::handler($client, $paramRef, $pageSetup);
}

sub handleLog {
	my ($client, $params, $callback, $httpClient, $response, $request) = @_;

	$response->header("Refresh" => "10; url=/" . $params->{path} . ($params->{lines} ? '?lines=' . $params->{lines} : '')) unless $params->{sorted};
	$response->header("Content-Type" => "text/plain; charset=utf-8");
		
	my $count = $params->{lines} || 50;

	my $body = '';
		
	if (-f $missing_log) {
		if ( $params->{sorted} && open(LOG, $missing_log) ) {
			$body = join('', sort {
				if ($a =~ /^#.*started/ || $b =~ /^#.*finished/) { -1 }
				elsif ($a =~ /^#.*finished/) { 1 }
				else { $a cmp $b }
			} <LOG>);
		}
		else {
			my $file = File::ReadBackwards->new( $missing_log );
				
			if ($file) {
		
				my @lines;
				while ( --$count && (my $line = $file->readline()) ) {
					unshift (@lines, $line);
				}
				$body .= join('', @lines);
		
				$file->close();			
			};		
		}
	}

	$response->header( 'Content-Length' => length($body) );
	
	return \$body;
}

1;